#include <time.h>
#include <unistd.h>
#include "timeout.h"
#include "scan.h"
#include "strerr.h"
#include "lexer.h"
#include "sml.h"
#include "optimize.h"
#include "stack.h"
#include "memlimit.h"

int
main(int argc, char **argv)
{
    unsigned long ul;
    long time_left;
    long elapsed_time;
    time_t start_time = time(0);
    int lookahead;

    if (argc != 3) strerr_die1x(1, "usage: press max-mem max-seconds");

    if (argv[1][scan_ulong(argv[1], &ul)] != '\000')
	strerr_die1x(1, "max-mem could not be parsed");
    memlimit(ul);

    if (argv[2][scan_ulong(argv[2], &ul)] != '\000')
	strerr_die1x(1, "max-seconds could not be parsed");
    time_left = (long) ul;

    yylex();

    elapsed_time = time(0) - start_time;
    if (elapsed_time >= time_left) {
	/*
	    We took longer than the allowed time just to read the
	    input, but they didn't kill us.  Hmm.  Just take another
	    3 seconds.
	*/
	time_left = 3;
    } else {
	time_left -= elapsed_time;
	if (time_left > 10) {
	    time_left -= 5;
	} else if (time_left > 1) {
	    time_left--;
	}
    }

    set_timeout(time_left);

    alloc_output();
    ContextStack_init();

    /* !!! consider incrementing lookahead by more than 1 */
    lookahead = 1;
    while (1)
    {
	optimize(lookahead);
	if (lookahead == markup.len - 1) break;
	if (lookahead * 2 >= markup.len) lookahead = markup.len - 1;
	else lookahead *= 2;
    }

    timeout();

    /*notreached*/
    return 0;
}

