#include <stdlib.h>
#include "stack.h"
#include "strerr.h"
#include "sml.h"
#include "timeout.h"

ContextStack *bottom;

typedef struct Block Block;

#define NumCells 50000

struct Block {
    Block *next;

    ContextStack cells[NumCells];
};

static Block *firstblock = 0;

static Block *currentblock;
static ContextStack *nextfree;
static ContextStack *blockend;

static void
reset(void)
{
    currentblock = firstblock;
    nextfree = bottom + 1;
    blockend = bottom + NumCells;
}

void
ContextStack_init(void)
{
    firstblock = malloc(sizeof *firstblock);
    if (!firstblock) quit();

    firstblock->next = 0;

    bottom = firstblock->cells;
    bottom->next = 0;
    bottom->x.flags = 0;
    bottom->x.size = 0;
    bottom->x.color = 0;
    bottom->x.tag = 0;

    reset();
}

static ContextStack *
getcell(void)
{
    if (nextfree == blockend) {
	if (!currentblock->next) {
	    currentblock->next = (Block *) malloc(sizeof(Block));
	    if (!currentblock->next) quit();
	    currentblock->next->next = 0;
	}

	currentblock = currentblock->next;
	nextfree = currentblock->cells;
	blockend = nextfree + NumCells;
    }

    return nextfree++;
}

ContextStack *
ContextStack_push(ContextStack *top, TagTypes tag)
{
    ContextStack *new_top = getcell();

    new_top->next = top;
    new_top->x = top->x;
    new_top->x.tag = tag;

    switch (tag) {
	case Tag_PL: new_top->x.flags = 0; break;
	case Tag_U: new_top->x.flags++; break;
	case Tag_B: new_top->x.flags |= B; break;
	case Tag_I: new_top->x.flags |= I; break;
	case Tag_TT: new_top->x.flags |= TT; break;
	case Tag_S: new_top->x.flags &= ~EM; new_top->x.flags |= S; break;
	case Tag_EM: new_top->x.flags |= EM; break;
	case Tag_0: case Tag_1: case Tag_2: case Tag_3: case Tag_4:
	case Tag_5: case Tag_6: case Tag_7: case Tag_8: case Tag_9:
	    new_top->x.size = '0' + tag - Tag_0; break;
	case Tag_r: new_top->x.color = 'r'; break;
	case Tag_g: new_top->x.color = 'g'; break;
	case Tag_b: new_top->x.color = 'b'; break;
	case Tag_w: new_top->x.color = 'w'; break;
	case Tag_c: new_top->x.color = 'c'; break;
	case Tag_m: new_top->x.color = 'm'; break;
	case Tag_y: new_top->x.color = 'y'; break;
	case Tag_k: new_top->x.color = 'k'; break;
	case Tag_Count: /* shut up gcc */ break;
    }

    return new_top;
}

static ContextStack *
gc_recurse(ContextStack *top)
{
    ContextStack *new_top;

    if (top == bottom) return top;

    top->next = gc_recurse(top->next);
    new_top = getcell();
    *new_top = *top;
    return new_top;
}

ContextStack *
ContextStack_gc(ContextStack *top)
{
    reset();
    return gc_recurse(top);
}

